@@ -2,6 +2,7 @@
 
 /* 
  * Copyright (C) 2001, 2002, 2003 Red Hat, Inc.
+ * Copyright (C) 2004, 2005 Elijah Newren
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -25,6 +26,12 @@
 #include "frame.h"
 #include "group.h"
 #include <X11/Xatom.h>
+#include <unistd.h>
+#include <string.h>
+#ifndef HOST_NAME_MAX
+/* Solaris headers apparently don't define this so do so manually; #326745 */
+#define HOST_NAME_MAX 255
+#endif
 
 typedef void (* InitValueFunc)   (MetaDisplay   *display,
                                   Atom           property,
@@ -159,7 +166,7 @@ reload_net_wm_pid (MetaWindow    *window,
       gulong cardinal = (int) value->v.cardinal;
       
       if (cardinal <= 0)
-        meta_warning (_("Application set a bogus _NET_WM_PID %ld\n"),
+        meta_warning (_("Application set a bogus _NET_WM_PID %lu\n"),
                       cardinal);
       else
         {
@@ -190,18 +197,77 @@ reload_net_wm_user_time (MetaWindow    *window,
     }
 }
 
+#define MAX_TITLE_LENGTH 512
+
+/**
+ * Called by set_window_title and set_icon_title to set the value of
+ * *target to title. It required and atom is set, it will update the
+ * appropriate property.
+ *
+ * Returns TRUE if a new title was set.
+ */
+static gboolean
+set_title_text (MetaWindow  *window,
+                gboolean     previous_was_modified,
+                const char  *title,
+                Atom         atom,
+                char       **target)
+{
+  char hostname[HOST_NAME_MAX + 1];
+  gboolean modified = FALSE;
+  
+  if (!target)
+    return FALSE;
+  
+  g_free (*target);
+  
+  if (!title)
+    *target = g_strdup ("");
+  else if (g_utf8_strlen (title, MAX_TITLE_LENGTH + 1) > MAX_TITLE_LENGTH)
+    {
+      *target = meta_g_utf8_strndup (title, MAX_TITLE_LENGTH);
+      modified = TRUE;
+    }
+  /* if WM_CLIENT_MACHINE indicates this machine is on a remote host
+   * lets place that hostname in the title */
+  else if (window->wm_client_machine &&
+           !gethostname (hostname, HOST_NAME_MAX + 1) &&
+           strcmp (hostname, window->wm_client_machine))
+    {
+      *target = g_strdup_printf (_("%s (on %s)"),
+                      title, window->wm_client_machine);
+      modified = TRUE;
+    }
+  else
+    *target = g_strdup (title);
+
+  if (modified && atom != None)
+    meta_prop_set_utf8_string_hint (window->display,
+                                    window->xwindow,
+                                    atom, *target);
+
+  /* Bug 330671 -- Don't forget to clear _NET_WM_VISIBLE_(ICON_)NAME */
+  if (!modified && previous_was_modified)
+    XDeleteProperty (window->display->xdisplay,
+                     window->xwindow,
+                     atom);
+
+  return modified;
+}
+
 static void
 set_window_title (MetaWindow *window,
                   const char *title)
 {
   char *str;
-  
-  g_free (window->title);
-  
-  if (title == NULL)
-    window->title = g_strdup ("");
-  else
-    window->title = g_strdup (title);
+ 
+  gboolean modified =
+    set_title_text (window,
+                    window->using_net_wm_visible_name,
+                    title,
+                    window->display->atom_net_wm_visible_name,
+                    &window->title);
+  window->using_net_wm_visible_name = modified;
   
   /* strndup is a hack since GNU libc has broken %.10s */
   str = g_strndup (window->title, 10);
@@ -246,8 +312,8 @@ reload_net_wm_name (MetaWindow    *window,
 
 static void
 init_wm_name (MetaDisplay   *display,
-                  Atom           property,
-                  MetaPropValue *value)
+              Atom           property,
+              MetaPropValue *value)
 {
   value->type = META_PROP_VALUE_TEXT_PROPERTY;
   value->atom = XA_WM_NAME;
@@ -281,12 +347,13 @@ static void
 set_icon_title (MetaWindow *window,
                 const char *title)
 {
-  g_free (window->icon_name);
-  
-  if (title == NULL)
-    window->icon_name = g_strdup ("");
-  else
-    window->icon_name = g_strdup (title);
+  gboolean modified =
+    set_title_text (window,
+                    window->using_net_wm_visible_icon_name,
+                    title,
+                    window->display->atom_net_wm_visible_icon_name,
+                    &window->icon_name);
+  window->using_net_wm_visible_icon_name = modified;
 }
 
 static void
@@ -351,6 +418,239 @@ reload_wm_icon_name (MetaWindow    *window,
     }
 }
 
+static void
+init_net_wm_state (MetaDisplay    *display,
+                   Atom            property,
+                   MetaPropValue  *value)
+{
+  value->type = META_PROP_VALUE_ATOM_LIST;
+  value->atom = display->atom_net_wm_state;
+}
+
+static void
+reload_net_wm_state (MetaWindow    *window,
+                     MetaPropValue *value)
+{
+  int i;
+
+  /* We know this is only an initial window creation,
+   * clients don't change the property.
+   */
+
+  window->shaded = FALSE;
+  window->maximized_horizontally = FALSE;
+  window->maximized_vertically = FALSE;
+  window->wm_state_modal = FALSE;
+  window->wm_state_skip_taskbar = FALSE;
+  window->wm_state_skip_pager = FALSE;
+  window->wm_state_above = FALSE;
+  window->wm_state_below = FALSE;
+  window->wm_state_demands_attention = FALSE;
+
+  if (value->type == META_PROP_VALUE_INVALID)
+    return;
+
+  i = 0;
+  while (i < value->v.atom_list.n_atoms)
+    {
+      if (value->v.atom_list.atoms[i] == window->display->atom_net_wm_state_shaded)
+        window->shaded = TRUE;
+      else if (value->v.atom_list.atoms[i] == window->display->atom_net_wm_state_maximized_horz)
+        window->maximize_horizontally_after_placement = TRUE;
+      else if (value->v.atom_list.atoms[i] == window->display->atom_net_wm_state_maximized_vert)
+        window->maximize_vertically_after_placement = TRUE;
+      else if (value->v.atom_list.atoms[i] == window->display->atom_net_wm_state_modal)
+        window->wm_state_modal = TRUE;
+      else if (value->v.atom_list.atoms[i] == window->display->atom_net_wm_state_skip_taskbar)
+        window->wm_state_skip_taskbar = TRUE;
+      else if (value->v.atom_list.atoms[i] == window->display->atom_net_wm_state_skip_pager)
+        window->wm_state_skip_pager = TRUE;
+      else if (value->v.atom_list.atoms[i] == window->display->atom_net_wm_state_fullscreen)
+        window->fullscreen = TRUE;
+      else if (value->v.atom_list.atoms[i] == window->display->atom_net_wm_state_above)
+        window->wm_state_above = TRUE;
+      else if (value->v.atom_list.atoms[i] == window->display->atom_net_wm_state_below)
+        window->wm_state_below = TRUE;
+      else if (value->v.atom_list.atoms[i] == window->display->atom_net_wm_state_demands_attention)
+        window->wm_state_demands_attention = TRUE;
+
+      ++i;
+    }
+
+  meta_verbose ("Reloaded _NET_WM_STATE for %s\n",
+                window->desc);
+
+  meta_window_recalc_window_type (window);
+}
+
+static void
+init_mwm_hints (MetaDisplay    *display,
+                Atom            property,
+                MetaPropValue  *value)
+{
+  value->type = META_PROP_VALUE_MOTIF_HINTS;
+  value->atom = display->atom_motif_wm_hints;
+}
+
+static void
+reload_mwm_hints (MetaWindow    *window,
+                  MetaPropValue *value)
+{
+  MotifWmHints *hints;
+
+  window->mwm_decorated = TRUE;
+  window->mwm_border_only = FALSE;
+  window->mwm_has_close_func = TRUE;
+  window->mwm_has_minimize_func = TRUE;
+  window->mwm_has_maximize_func = TRUE;
+  window->mwm_has_move_func = TRUE;
+  window->mwm_has_resize_func = TRUE;
+
+  if (value->type == META_PROP_VALUE_INVALID)
+    {
+      meta_verbose ("Window %s has no MWM hints\n", window->desc);
+      meta_window_recalc_features (window);
+      return;
+    }
+
+  hints = value->v.motif_hints;
+
+  /* We support those MWM hints deemed non-stupid */
+
+  meta_verbose ("Window %s has MWM hints\n",
+                window->desc);
+
+  if (hints->flags & MWM_HINTS_DECORATIONS)
+    {
+      meta_verbose ("Window %s sets MWM_HINTS_DECORATIONS 0x%lx\n",
+          window->desc, hints->decorations);
+
+      if (hints->decorations == 0)
+        window->mwm_decorated = FALSE;
+      /* some input methods use this */
+      else if (hints->decorations == MWM_DECOR_BORDER)
+        window->mwm_border_only = TRUE;
+    }
+  else
+    meta_verbose ("Decorations flag unset\n");
+
+  if (hints->flags & MWM_HINTS_FUNCTIONS)
+    {
+      gboolean toggle_value;
+
+      meta_verbose ("Window %s sets MWM_HINTS_FUNCTIONS 0x%lx\n",
+                    window->desc, hints->functions);
+
+      /* If _ALL is specified, then other flags indicate what to turn off;
+       * if ALL is not specified, flags are what to turn on.
+       * at least, I think so
+       */
+
+      if ((hints->functions & MWM_FUNC_ALL) == 0)
+        {
+          toggle_value = TRUE;
+
+          meta_verbose ("Window %s disables all funcs then reenables some\n",
+                        window->desc);
+          window->mwm_has_close_func = FALSE;
+          window->mwm_has_minimize_func = FALSE;
+          window->mwm_has_maximize_func = FALSE;
+          window->mwm_has_move_func = FALSE;
+          window->mwm_has_resize_func = FALSE;
+        }
+      else
+        {
+          meta_verbose ("Window %s enables all funcs then disables some\n",
+                        window->desc);
+          toggle_value = FALSE;
+        }
+
+      if ((hints->functions & MWM_FUNC_CLOSE) != 0)
+        {
+          meta_verbose ("Window %s toggles close via MWM hints\n",
+                        window->desc);
+          window->mwm_has_close_func = toggle_value;
+        }
+      if ((hints->functions & MWM_FUNC_MINIMIZE) != 0)
+        {
+          meta_verbose ("Window %s toggles minimize via MWM hints\n",
+                        window->desc);
+          window->mwm_has_minimize_func = toggle_value;
+        }
+      if ((hints->functions & MWM_FUNC_MAXIMIZE) != 0)
+        {
+          meta_verbose ("Window %s toggles maximize via MWM hints\n",
+                        window->desc);
+          window->mwm_has_maximize_func = toggle_value;
+        }
+      if ((hints->functions & MWM_FUNC_MOVE) != 0)
+        {
+          meta_verbose ("Window %s toggles move via MWM hints\n",
+                        window->desc);
+          window->mwm_has_move_func = toggle_value;
+        }
+      if ((hints->functions & MWM_FUNC_RESIZE) != 0)
+        {
+          meta_verbose ("Window %s toggles resize via MWM hints\n",
+                        window->desc);
+          window->mwm_has_resize_func = toggle_value;
+        }
+    }
+  else
+    meta_verbose ("Functions flag unset\n");
+
+  meta_window_recalc_features (window);
+  
+  /* We do all this anyhow at the end of meta_window_new() */
+  if (!window->constructing)
+    {
+      if (window->decorated)
+        meta_window_ensure_frame (window);
+      else
+        meta_window_destroy_frame (window);
+      
+      meta_window_queue_move_resize (window);
+      /* because ensure/destroy frame may unmap */
+      meta_window_queue_calc_showing (window);
+    }
+}
+
+static void
+init_wm_class (MetaDisplay    *display,
+               Atom            property,
+               MetaPropValue  *value)
+{
+  value->type = META_PROP_VALUE_CLASS_HINT;
+  value->atom = XA_WM_CLASS;
+}
+
+static void
+reload_wm_class (MetaWindow    *window,
+                 MetaPropValue *value)
+{
+  if (window->res_class)
+    g_free (window->res_class);
+  if (window->res_name)
+    g_free (window->res_name);
+
+  window->res_class = NULL;
+  window->res_name = NULL;
+
+  if (value->type != META_PROP_VALUE_INVALID)
+    { 
+      if (value->v.class_hint.res_name)
+        window->res_name = g_strdup (value->v.class_hint.res_name);
+
+      if (value->v.class_hint.res_class)
+        window->res_class = g_strdup (value->v.class_hint.res_class);
+    }
+
+  meta_verbose ("Window %s class: '%s' name: '%s'\n",
+      window->desc,
+      window->res_class ? window->res_class : "none",
+      window->res_name ? window->res_name : "none");
+}
+
 static void
 init_net_wm_desktop (MetaDisplay   *display,
                      Atom           property,
@@ -832,9 +1132,64 @@ reload_wm_hints (MetaWindow    *window,
   meta_window_queue_move_resize (window);  
 }
 
+static void
+init_transient_for (MetaDisplay   *display,
+                    Atom           property,
+                    MetaPropValue *value)
+{
+  value->type = META_PROP_VALUE_WINDOW;
+  value->atom = XA_WM_TRANSIENT_FOR;
+}
 
+static void
+reload_transient_for (MetaWindow    *window,
+                      MetaPropValue *value)
+{
+  window->xtransient_for = None;
+  
+  if (value->type != META_PROP_VALUE_INVALID)
+    window->xtransient_for = value->v.xwindow;
 
-#define N_HOOKS 24
+  /* Make sure transient_for is valid */
+  if (window->xtransient_for != None &&
+      meta_display_lookup_x_window (window->display, 
+                                    window->xtransient_for) == NULL)
+    {
+      window->xtransient_for = None;
+      meta_warning (_("Invalid WM_TRANSIENT_FOR window 0x%lx specified "
+                      "for %s.\n"),
+                    window->xtransient_for, window->desc);
+    }
+
+  window->transient_parent_is_root_window =
+    window->xtransient_for == window->screen->xroot;
+
+  if (window->xtransient_for != None)
+    meta_verbose ("Window %s transient for 0x%lx (root = %d)\n", window->desc,
+        window->xtransient_for, window->transient_parent_is_root_window);
+  else
+    meta_verbose ("Window %s is not transient\n", window->desc);
+
+  /* may now be a dialog */
+  meta_window_recalc_window_type (window);
+
+  /* update stacking constraints */
+  meta_stack_update_transient (window->screen->stack, window);
+
+  /* possibly change its group. We treat being a window's transient as
+   * equivalent to making it your group leader, to work around shortcomings
+   * in programs such as xmms-- see #328211.
+   */
+  if (window->xtransient_for != None &&
+      window->xgroup_leader != None &&
+      window->xtransient_for != window->xgroup_leader)
+    meta_window_group_leader_changed (window);
+
+  if (!window->constructing)
+    meta_window_queue_move_resize (window);
+}
+
+#define N_HOOKS 25
 
 void
 meta_display_init_window_prop_hooks (MetaDisplay *display)
@@ -890,13 +1245,13 @@ meta_display_init_window_prop_hooks (MetaDisplay *display)
   ++i;
 
   hooks[i].property = display->atom_net_wm_state;
-  hooks[i].init_func = NULL;
-  hooks[i].reload_func = NULL;
+  hooks[i].init_func = init_net_wm_state;
+  hooks[i].reload_func = reload_net_wm_state;
   ++i;
   
   hooks[i].property = display->atom_motif_wm_hints;
-  hooks[i].init_func = NULL;
-  hooks[i].reload_func = NULL;
+  hooks[i].init_func = init_mwm_hints;
+  hooks[i].reload_func = reload_mwm_hints;
   ++i;
 
   hooks[i].property = display->atom_net_wm_icon_geometry;
@@ -905,8 +1260,8 @@ meta_display_init_window_prop_hooks (MetaDisplay *display)
   ++i;
 
   hooks[i].property = XA_WM_CLASS;
-  hooks[i].init_func = NULL;
-  hooks[i].reload_func = NULL;
+  hooks[i].init_func = init_wm_class;
+  hooks[i].reload_func = reload_wm_class;
   ++i;
 
   hooks[i].property = display->atom_wm_client_leader;
@@ -969,6 +1324,11 @@ meta_display_init_window_prop_hooks (MetaDisplay *display)
   hooks[i].reload_func = reload_wm_hints;
   ++i;
   
+  hooks[i].property = XA_WM_TRANSIENT_FOR;
+  hooks[i].init_func = init_transient_for;
+  hooks[i].reload_func = reload_transient_for;
+  ++i;
+
   if (i != N_HOOKS)
     g_error ("Initialized %d hooks should have been %d\n", i, N_HOOKS);
 }
