@@ -4,7 +4,7 @@
  * Copyright (C) 2001 Havoc Pennington
  * Copyright (C) 2002 Red Hat, Inc.
  * Copyright (C) 2003, 2004 Rob Adams
- * Copyright (C) 2004, 2005 Elijah Newren
+ * Copyright (C) 2004-2006 Elijah Newren
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -58,6 +58,13 @@ typedef enum
   META_MAXIMIZE_VERTICAL   = 1 << 1
 } MetaMaximizeFlags;
 
+typedef enum {
+  META_CLIENT_TYPE_UNKNOWN = 0,
+  META_CLIENT_TYPE_APPLICATION = 1,
+  META_CLIENT_TYPE_PAGER = 2,
+  META_CLIENT_TYPE_MAX_RECOGNIZED = 2
+} MetaClientType;
+
 struct _MetaStruts
 {
   /* struts */
@@ -139,6 +146,7 @@ struct _MetaWindow
 
   /* Minimize is the state controlled by the minimize button */
   guint minimized : 1;
+  guint was_minimized : 1;
   guint tab_unminimized : 1;
 
   /* Whether the window is mapped; actual server-side state
@@ -232,6 +240,9 @@ struct _MetaWindow
   /* Are we in meta_window_free()? */
   guint unmanaging : 1;
 
+  /* Are we in meta_window_new()? */
+  guint constructing : 1;
+  
   /* Are we in the calc_showing queue? */
   guint calc_showing_queued : 1;
 
@@ -263,8 +274,10 @@ struct _MetaWindow
   guint transient_parent_is_root_window : 1;
 
   /* Info on which props we got our attributes from */
-  guint using_net_wm_name : 1; /* vs. plain wm_name */
-  guint using_net_wm_icon_name : 1; /* vs. plain wm_icon_name */
+  guint using_net_wm_name              : 1; /* vs. plain wm_name */
+  guint using_net_wm_visible_name      : 1; /* tracked so we can clear it */
+  guint using_net_wm_icon_name         : 1; /* vs. plain wm_icon_name */
+  guint using_net_wm_visible_icon_name : 1; /* tracked so we can clear it */
 
   /* has a shape mask */
   guint has_shape : 1;
@@ -419,6 +432,9 @@ gboolean    meta_window_showing_on_its_workspace (MetaWindow *window);
 /* Return whether the window should be currently mapped */
 gboolean    meta_window_should_be_showing   (MetaWindow  *window);
 
+/* See warning in window.c about this function */
+gboolean    __window_is_terminal (MetaWindow *window);
+
 /* This recalcs the window/frame size, and recalcs the frame
  * size/contents as well.
  */
@@ -449,8 +465,8 @@ void        meta_window_get_geometry         (MetaWindow  *window,
                                               int         *y,
                                               int         *width,
                                               int         *height);
-void        meta_window_get_outer_rect       (MetaWindow    *window,
-                                              MetaRectangle *rect);
+void        meta_window_get_outer_rect       (const MetaWindow *window,
+                                              MetaRectangle    *rect);
 void        meta_window_get_xor_rect         (MetaWindow          *window,
                                               const MetaRectangle *grab_wireframe_rect,
                                               MetaRectangle       *xor_rect);
@@ -529,6 +545,8 @@ gboolean meta_window_same_application (MetaWindow *window,
   (((w)->input || (w)->take_focus ) && META_WINDOW_IN_NORMAL_TAB_CHAIN_TYPE (w) && (!(w)->skip_taskbar))
 #define META_WINDOW_IN_DOCK_TAB_CHAIN(w) \
   (((w)->input || (w)->take_focus) && ! META_WINDOW_IN_NORMAL_TAB_CHAIN_TYPE (w))
+#define META_WINDOW_IN_GROUP_TAB_CHAIN(w, g) \
+  (((w)->input || (w)->take_focus) && (!g || meta_window_get_group(w)==g))
 
 void meta_window_refresh_resize_popup (MetaWindow *window);
 
@@ -542,6 +560,8 @@ gboolean meta_window_is_ancestor_of_transient (MetaWindow            *window,
 void     meta_window_foreach_ancestor         (MetaWindow            *window,
                                                MetaWindowForeachFunc  func,
                                                void                  *data);
+MetaWindow* meta_window_find_root_ancestor    (MetaWindow *window);
+
 
 void meta_window_begin_grab_op (MetaWindow *window,
                                 MetaGrabOp  op,
@@ -558,7 +578,8 @@ gboolean meta_window_get_icon_geometry (MetaWindow    *window,
 
 const char* meta_window_get_startup_id (MetaWindow *window);
 
-void meta_window_recalc_features (MetaWindow *window);
+void meta_window_recalc_features    (MetaWindow *window);
+void meta_window_recalc_window_type (MetaWindow *window);
 
 void meta_window_queue_update_icon (MetaWindow *window);
 
@@ -568,4 +589,7 @@ void meta_window_stack_just_below (MetaWindow *window,
 void meta_window_set_user_time (MetaWindow *window,
                                 Time        timestamp);
 
+void meta_window_set_demands_attention (MetaWindow *window);
+
+void meta_window_unset_demands_attention (MetaWindow *window);
 #endif
